import processing.serial.*;
import processing.video.*;
import beads.*;
import java.io.*;

boolean RECORD = false;
boolean REPLAY = true;
boolean SAVE_VIDEO = false; // disables audio
boolean SAVE_AUDIO = false;  // disables video

// Audio system
AudioContext ac;
float frequencies[][] = {{220, 233.08, 246.96, 261.64},
                         {277.20, 293.68, 311.12, 329.64},
                         {349.24, 370.00, 392.00, 415.32},
                         {440.0, 466.16, 493.92, 523.28}};
WavePlayer waves[][] = new WavePlayer[4][4];
Gain gains[][] = new Gain[4][4];

// Object to record screen
PrintWriter output;
MovieMaker mm;
RecordToFile recorder;
BufferedReader input;

Serial myPort;  // Create object from Serial class
int RUNNING_AVERAGE_SIZE = 5;

int[][][] readings = new int[4][4][RUNNING_AVERAGE_SIZE];
float[][] panel_heights = new float[4][4];
int average_pos = 0;

float time = 0;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void setup()
{
  size(1000, 600, P3D);
  if (!REPLAY) {
    String portName = Serial.list()[0];
    println(portName);
    myPort = new Serial(this, portName, 115200); 
    myPort.clear();
  }
  smooth();
  
  if (SAVE_VIDEO && REPLAY)
    frameRate(240);
  else
    frameRate(60);
  
  ac = new AudioContext();
  
  if (SAVE_VIDEO) {
    mm = new MovieMaker(this, width, height, "drawing.avi", 60);
  }
  
  if (SAVE_AUDIO) {
    try { 
      recorder = new RecordToFile(ac, 1, new File("/Users/mkeeter/grad/classes/MAS.863/multitouch/rev1b/viz/multitouch_hear/audio.wav"));
    } catch(IOException e) { 
      e.printStackTrace(); 
    }
  }
  
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      waves[i][j] = new WavePlayer(ac, frequencies[i][3-j], Buffer.SINE);
      gains[i][j] = new Gain(ac, 1, 0.0);
      gains[i][j].addInput(waves[i][j]);
      ac.out.addInput(gains[i][j]);
      if (SAVE_AUDIO)
        recorder.addInput(gains[i][j]);
    }
  }

  if (SAVE_AUDIO)
      ac.out.addDependent(recorder);
  
  if (RECORD)
    output = createWriter("recording.txt");
   
  if (!SAVE_VIDEO)
    ac.start();
    
  if (REPLAY)
    input = createReader("recording.txt");
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void process_pad_data()
{
  float min_panel = 1000;
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      for (int k = 0; k < RUNNING_AVERAGE_SIZE; ++k)
        panel_heights[i][j] += readings[3-i][j][k];
      panel_heights[i][j] /= float(RUNNING_AVERAGE_SIZE);
      panel_heights[i][j] = (70 - panel_heights[i][j])  / 40.;
      if (panel_heights[i][j] < min_panel)
        min_panel = panel_heights[i][j];
    }
  }
  
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      panel_heights[i][j] -= min_panel;
      if (RECORD) {
        output.print(panel_heights[i][j]);
        output.print(' ');
      }
    }
  }
  if (RECORD)
    output.print('\n');

  if (++average_pos == RUNNING_AVERAGE_SIZE)
    average_pos = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void draw_pad_data()
{
  stroke(128);
  pushMatrix();
  translate(19*width/20., height - width/20., 0);
  scale(width/40. * 0.9, width/40. * 0.9);
  
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      float panel_height = panel_heights[i][j];

      pushMatrix();
      translate(i - 1.5, j - 1.5);
        
      fill(255, (panel_height - 0.4) / 0.4 * 255.0);

      beginShape(QUADS);
      vertex(-0.5, -0.5);
      vertex( 0.5, -0.5);
      vertex( 0.5,  0.5);
      vertex(-0.5,  0.5);
      endShape();
      
      popMatrix();
    }
  }
  popMatrix();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
float adjust_volumes()
{
  float total_gain = 0, scale_factor = 1;
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      total_gain += min(1, max(0, panel_heights[i][j] - 0.2));
    }
  }
  
  if (total_gain > 1)
    scale_factor = 1 / total_gain;
    
  for (int i = 0; i < 4; ++i) {
    for (int j = 0; j < 4; ++j) {
      float audio_volume = min(1, max(0, panel_heights[i][j] - 0.2)) * scale_factor;
      if (audio_volume == 0)
        gains[i][j].pause(true);
      else
        gains[i][j].pause(false);
      if (SAVE_AUDIO)
        gains[i][j].setGain(audio_volume/2.0);
      else
        gains[i][j].setGain(audio_volume);
    }
  }
  return scale_factor;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void drawSine(float freq, float time, float magnitude, float h)
{
  int dx = 1;
  stroke(h, 255, 255);
  
  for(int x = 0; x < width-1; x += dx) {
    float window = sin(PI * x / width) * magnitude;
    float y1 = window * sin( (width + x + time) * freq / width) * height/2.0 + height / 2.0;
    float y2 = window * sin( (width + x + dx + time) * freq / width) * height/2.0 + height / 2.0;
    line(x, y1, x + dx, y2);
  }

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void read_serial_data()
{
  int i = 0, j = 0;

  myPort.clear();

  // Wait we get a sync frame
  do
  {
    while(myPort.available() == 0);
  } while (myPort.readChar() != '\n');

   
  while (j < 4) {
    while(myPort.available() == 0);
    readings[i][j][average_pos] = myPort.read();
    if (++i == 4) {
      i = 0;
      j++;
    }
  }
}  

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int read_from_file()
{
  String in;
  try {
    in = input.readLine();
  } catch (IOException e) {
    e.printStackTrace();
    in = null;
  }
  
  if (in == null)
    return 1;
    
  float values[] = float(split(in, ' '));
  int x = 0;
  for (int i = 0; i < 4; ++i)
    for (int j = 0; j < 4; ++j)
      panel_heights[i][j] = values[x++];

  return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void draw()
{
  background(0);
  noFill();
  
  if (!REPLAY) {
    read_serial_data();
    process_pad_data();
  } else {
    if (read_from_file() == 1)
      shutdown();
  }
  
  float scale_factor =  adjust_volumes();
  
  if (!SAVE_AUDIO) {
    draw_pad_data();
  
    colorMode(HSB);
    for (int i = 0; i < 4; ++i) {
        for (int j = 0; j < 4; ++j) {
          float power = gains[i][j].getGain() / scale_factor;
          drawSine(frequencies[i][3-j], time, power, 255.0 * (i * 4 + j) / 16.0);
        }
    }
    colorMode(RGB);
    time += 4;
  }
   
  if (SAVE_VIDEO)
    mm.addFrame();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void keyPressed()
{
  if (key == 'q')
    shutdown();
}

void shutdown()
{
  if (RECORD) {
    output.flush(); // Write the remaining data
    output.close(); // Finish the file
  }
  if (SAVE_AUDIO)
    recorder.kill();
  if (SAVE_VIDEO)
    mm.finish();
    
  ac.stop();
  exit();
}
      
